'''Run all dossier.fc tests.

.. This software is released under an MIT/X11 open source license.
   Copyright 2012-2014 Diffeo, Inc.

'''
from __future__ import absolute_import, division, print_function
import argparse
import os

try:
    import pytest
except ImportError:
    pytest = None

import dossier.fc.tests.performance as performance_tests


def main():
    parser = argparse.ArgumentParser(
        description='Run all dossier.fc tests.',
    )
    if pytest:
        parser.add_argument('--unit', action='store_true',
                            help='Run unit tests')
    parser.add_argument('--performance', '--perf', action='store_true',
                        help='Run performance tests')
    args = parser.parse_args()

    if not args.unit and not args.performance:
        args.unit = (pytest is not None)
        args.performance = True

    if args.unit:
        pytest.main(os.path.dirname(__file__))

    if args.performance:
        performance_tests.main()


if __name__ == '__main__':
    main()



'''Serializable feature-like thing for tests.

.. This software is released under an MIT/X11 open source license.
   Copyright 2012-2014 Diffeo, Inc.

'''
from __future__ import absolute_import, division, print_function
import json


class Thing(object):
    def __init__(self, blob=None):
        self.data = dict()
        if blob is not None:
            self.loads(blob)

    def dumps(self):
        return json.dumps(self.data)

    def loads(self, blob):
        self.data = json.loads(blob)

    def __getitem__(self, key):
        return self.data[key]

    def __setitem__(self, key, value):
        self.data[key] = value

    def do_more_things(self):
        self.data['doing'] = 'something'

    def __eq__(self, other):
        return self.data == other.data

    def __repr__(self):
        return 'Thing({!r})'.format(self.dumps())


class ThingSerializer(object):
    def __init__(self):
        raise NotImplementedError

    @staticmethod
    def loads(blob):
        return Thing(blob)

    @staticmethod
    def dumps(thing):
        return thing.dumps()

    constructor = Thing



'''dossier.fc Feature Collections

.. This software is released under an MIT/X11 open source license.
   Copyright 2012-2014 Diffeo, Inc.

Tests for counters which require support for `__missing__` in
FeatureCollection.
'''
from __future__ import absolute_import, division, print_function
from collections import Counter

import pytest

from dossier.fc import FeatureCollection, StringCounter, SerializationError
from dossier.fc.tests import counter_type


def assert_same_fc(fc1, fc2):
    '''Verify that fc1 and fc2 have the same data, otherwise raise'''

    ## verify the keys have not changed
    keys1 = set(fc1.keys())
    keys2 = set(fc2.keys())
    assert keys1 == keys2, keys2 - keys1

    ## verify the vals ahve not changed
    vals1 = set()
    for mset in fc1.values():
        map(vals1.add, mset.items())
    vals2 = set()
    for mset in fc2.values():
        map(vals2.add, mset.items())
    assert vals1 == vals2, vals1 - vals2


def test_bundle_default(counter_type):
    'does it make a new counter that adds properly'
    fc = FeatureCollection()
    assert isinstance(fc['foo'], counter_type)

    fc['foo'] += counter_type(Counter('dog'))
    assert isinstance(fc['foo'], counter_type), \
        'failed and made %s' % type(fc['foo'])

    fc['foo'] -= counter_type(Counter('dog'))
    assert isinstance(fc['foo'], counter_type), \
        'failed and made %s' % type(fc['foo'])

    if hasattr(fc['foo'], 'substract'):
        fc['foo'].subtract(counter_type(Counter('dog')))
        assert isinstance(fc['foo'], counter_type), \
            'failed and made %s' % type(fc['foo'])
        fc['foo'] += counter_type(Counter('dog'))

    fc['foo'] += counter_type(Counter('dog'))
    assert isinstance(fc['foo'], counter_type), \
        'failed and made %s' % type(fc['foo'])
    fc['foo'] += counter_type(Counter('dog'))
    fc['foo'] += counter_type(Counter('dog cat'))
    assert Counter(fc['foo'].values()) == Counter({1: 4, 3: 3})


def test_fc_build_from_dict(counter_type):
    fc = FeatureCollection({
            'hello': counter_type(Counter('hello')), 
            'goodbye': counter_type(Counter('goodbye'))})
    assert Counter(fc['hello'].values()) == Counter({1: 3, 2: 1})
    assert isinstance(fc['hello'], counter_type)


def test_fc_meta_adding(counter_type):
    fc = FeatureCollection({
            'hello': counter_type(Counter('hello')),
            'goodbye': counter_type(Counter('goodbye'))})
    fc2 = fc + fc
    assert Counter(fc2['hello'].values()) == Counter({2: 3, 4: 1})


def test_fc_eq(counter_type):
    fc1 = FeatureCollection({
            'hello': counter_type(Counter('hello')),
            'goodbye': counter_type(Counter('goodbye'))})
    fc2 = FeatureCollection({
            'hello': counter_type(Counter('hello')),
            'goodbye': counter_type(Counter('goodbye'))})
    fc3 = FeatureCollection({
            'hello': counter_type(Counter('hello')),
            'goodbye': counter_type(Counter('goodbye2'))})

    assert fc1 == fc2
    assert fc1 != fc3
    

def test_fc_meta_adding_complex(counter_type):
    fc = FeatureCollection({
            'hello': counter_type(Counter('hello')), 
            'goodbye': counter_type(Counter('goodbye'))})
    fc2 = FeatureCollection({
            'hello': counter_type(Counter('hello')),
            'goodbye': counter_type(Counter('goodbye'))})
    fc3 = fc + fc2

    assert Counter(fc3['hello'].values()) == Counter({2: 3, 4: 1})
    fc += fc2
    assert Counter(fc['hello'].values()) == Counter({2: 3, 4: 1})

    fc3 -= fc2
    assert Counter(fc3['hello'].values()) == Counter({1: 3, 2: 1})
    
    fc3 -= fc2
    assert Counter(fc3['hello'].values()) == Counter()


def test_string_counter():
    hc = StringCounter('this sentence has been parsed')
    h1 = (id(hc), hc.generation)

    cache = dict(h1='hi')
    assert len(cache) == 1

    hc.update('more text')
    h2 = (id(hc), hc.generation)

    cache[h2] = 'hi again?'
    assert len(cache) == 2
    assert h1 != h2


def test_string_counter2():
    sc = StringCounter()
    sc['hi'] += 1
    assert isinstance(sc, StringCounter)
    sc += Counter('hi')
    assert isinstance(sc, StringCounter)


def test_values(counter_type):
    c = counter_type(Counter('dog'))
    assert set(map(int, c.values())) == set([1])
    assert len(c.values()) == 3


def test_entity(counter_type):
    ## build entity, serialize, deserialize, and verify its multisets
    fc1 = FeatureCollection()
    fc1['bow'] += counter_type(Counter(['big', 'dog']))
    fc1['bow'] += counter_type(Counter('tall building'))
    fc1['bon'] += counter_type(Counter(['Super Cat', 'Small Cat', 'Tiger Fish']))

    ## there should be nine items of size 1
    assert Counter(fc1['bow'].values())[1] == 10, fc1['bow'].items()

    ## double the counts, should recurse down
    fc1 += fc1

    ## check values doubled
    assert Counter(fc1['bow'].values())[2] == 10, fc1['bow'].items()

    ## serialize/deserialize it
    blob = fc1.dumps()
    assert_same_fc(fc1, FeatureCollection.loads(blob))

    ## deserialize it via chunk
    fc2 = FeatureCollection.loads(fc1.dumps())
    assert_same_fc(fc1, fc2)


def test_multiset_equality(counter_type):
    ent1 = FeatureCollection()
    ent2 = FeatureCollection()
    ent1['bow'] += counter_type(Counter(['big', 'dog']))
    ent2['bow'] += counter_type(Counter(['big', 'dog']))
    assert ent1 == ent2


def test_multiset_change(counter_type):
    ent1 = FeatureCollection()
    ent1['bow'] += counter_type(Counter(['big', 'dog']))
    ent1.pop('bow')
    assert dict(ent1.items()) == dict()

    ## can pop empty -- fails
    #ent1.pop('foo')

    ## set equal to
    test_data = ['big2', 'dog2']
    ent1['bow'] = counter_type(Counter(test_data))
    assert list(ent1['bow'].values()) == [1,1]

    ent1['bow'] += counter_type(Counter(test_data))
    assert list(ent1['bow'].values()) == [2,2]


def test_serialize_deserialize(counter_type):
    ## build entity, serialize, deserialize, and verify its multisets
    ent1 = FeatureCollection()
    ent1['bow'] += counter_type(Counter(['big', 'dog']))
    ent1['bow'] += counter_type(Counter('tall building'))
    ent1['bon'] += counter_type(Counter(['Super Cat', 'Small Cat',
                                         'Tiger Fish']))

    blob = ent1.dumps()
    ent2 = FeatureCollection.loads(blob)
    assert_same_fc(ent1, ent2)


def test_type(counter_type):
    ent1 = FeatureCollection()
    ent1['bow'] += counter_type(Counter(['big', 'dog']))
    if counter_type.__name__ == 'StringCounter':
        ent1['bow']['a'] += 1
    assert isinstance(ent1, FeatureCollection)

    ent3 = FeatureCollection()
    ent3['bow'] += counter_type(Counter(['cat']))

    ent1 += ent3
    assert isinstance(ent1, FeatureCollection)


def test_clone(counter_type):
    ent1 = FeatureCollection()
    ent1['bow'] += counter_type(Counter(['cat']))
    ent2 = FeatureCollection(ent1)
    assert ent2 is not ent1
    assert ent2 == ent1


def test_non_counter_features_serialize():
    fc = FeatureCollection({'NAME': u'foobaz'})
    fc = FeatureCollection.loads(fc.dumps())


def test_non_counter_features_bad_serialize():
    with pytest.raises(SerializationError):
        FeatureCollection({'NAME': 'foobaz'})
    fc = FeatureCollection()
    fc['NAME'] = 'foobaz'
    with pytest.raises(SerializationError):
        fc.dumps()


def test_non_counter_features_total():
    fc = FeatureCollection({'NAME': u'foobaz'})
    assert fc.total() == 0


def test_non_counter_features_inequality():
    fc1 = FeatureCollection({'NAME': u'foobaz'})
    fc2 = FeatureCollection({'NAME': u'blahblahblah'})
    assert fc1 != fc2


def test_binop_no_share():
    fc1 = FeatureCollection({'NAME': {'foo': 1, 'bar': 1}})
    fc2 = FeatureCollection({'NAME': {'foo': 2, 'bar': 2}})

    fc3 = fc1 + fc2

    assert fc1['NAME']['foo'] == 1
    assert fc2['NAME']['foo'] == 2

    fc1 += fc2
    assert fc1 == fc3
    assert fc1['NAME']['foo'] == 3
    assert fc2['NAME']['foo'] == 2


def test_binop_different_no_share():
    fc1 = FeatureCollection({'FOO': {'foo': 1}})
    fc2 = FeatureCollection({'BAR': {'bar': 1}})

    result = fc1 + fc2
    expected = FeatureCollection({'FOO': {'foo': 1 }, 'BAR': {'bar': 1}})
    assert result == expected

    result['BAR']['bar'] = 2
    assert fc2['BAR']['bar'] == 1

    result['FOO']['foo'] = 2
    assert fc1['FOO']['foo'] == 1


def test_subtraction():
    fc1 = FeatureCollection({'a': {'a': 3, 'b': 2}, 'b': {'b': 2}})
    fc2 = FeatureCollection({'a': {'a': 3, 'c': 1}, 'c': {'c': 1}})
    fc3 = fc1 - fc2

    assert fc3 == FeatureCollection({'a': {'b': 2}, 'b': {'b': 2}, 'c': {}})



'''dossier.fc Feature Collections

.. This software is released under an MIT/X11 open source license.
   Copyright 2012-2014 Diffeo, Inc.

tests for FeatureCollection which has __missing__ that detects which kind
of counter is active.
'''
from __future__ import absolute_import, division, print_function
from collections import Counter

from dossier.fc import FeatureCollection, StringCounter

from dossier.fc.tests import counter_type


def test_default(counter_type):
    'does a FC make a new counter that adds properly'
    mc = FeatureCollection()
    assert isinstance(mc['foo'], counter_type)

    mc['foo'] += counter_type(Counter('dog'))
    assert isinstance(mc['foo'], counter_type), \
        'failed and made %s' % type(mc['foo'])

    mc['foo'] -= counter_type(Counter('dog'))
    assert isinstance(mc['foo'], counter_type), \
        'failed and made %s' % type(mc['foo'])

    if hasattr(mc['foo'], 'substract'):
        mc['foo'].subtract(counter_type(Counter('dog')))
        assert isinstance(mc['foo'], counter_type), \
            'failed and made %s' % type(mc['foo'])
        mc['foo'] += counter_type(Counter('dog'))

    mc['foo'] += counter_type(Counter('dog'))
    assert isinstance(mc['foo'], counter_type), \
        'failed and made %s' % type(mc['foo'])

    mc['foo'] += counter_type(Counter('dog'))
    mc['foo'] += counter_type(Counter('dog cat'))
    assert Counter(mc['foo'].values()) == Counter({1: 4, 3: 3})


def test_build_from_dict(counter_type):
    mc = FeatureCollection({
            'hello': counter_type(Counter('hello')), 
            'goodbye': counter_type(Counter('goodbye'))})

    assert Counter(mc['hello'].values()) == Counter({1: 3, 2: 1})
    assert isinstance(mc['hello'], counter_type)


def test_meta_adding(counter_type):
    mc = FeatureCollection({
            'hello': counter_type(Counter('hello')),
            'goodbye': counter_type(Counter('goodbye'))})
    mc2 = mc + mc

    assert Counter(mc2['hello'].values()) == Counter({2: 3, 4: 1})


def test_eq(counter_type):
    mc1 = FeatureCollection({
            'hello': counter_type(Counter('hello')),
            'goodbye': counter_type(Counter('goodbye'))})
    mc2 = FeatureCollection({
            'hello': counter_type(Counter('hello')),
            'goodbye': counter_type(Counter('goodbye'))})
    mc3 = FeatureCollection({
            'hello': counter_type(Counter('hello')),
            'goodbye': counter_type(Counter('goodbye2'))})

    assert mc1 == mc2
    assert mc1 != mc3
    

def test_meta_adding_complex(counter_type):
    mc = FeatureCollection({
            'hello': counter_type(Counter('hello')), 
            'goodbye': counter_type(Counter('goodbye'))})
    mc2 = FeatureCollection({
            'hello': counter_type(Counter('hello')),
            'goodbye': counter_type(Counter('goodbye'))})
    mc3 = mc + mc2

    assert Counter(mc3['hello'].values()) == Counter({2: 3, 4: 1})
    mc += mc2
    assert Counter(mc['hello'].values()) == Counter({2: 3, 4: 1})

    ## isub tests
    mc3 -= mc2
    assert Counter(mc3['hello'].values()) == Counter({1: 3, 2: 1})
    
    mc3 -= mc2
    assert Counter(mc3['hello'].values()) == Counter()


def test_type(counter_type):
    m1 = FeatureCollection()
    m1['bow'] += counter_type(Counter(['big', 'dog']))

    assert type(m1) == FeatureCollection

    m2 = FeatureCollection()
    m2['bow'] += counter_type(Counter(['cat']))
    m1 += m2

    assert type(m1) == FeatureCollection


def test_string_counter():
    hc = StringCounter('this sentence has been parsed')
    h1 = (id(hc), hc.generation)

    cache = dict(h1='hi')
    assert len(cache) == 1

    hc.update('more text')
    h2 = (id(hc), hc.generation)
    cache[h2] = 'hi again?'

    assert len(cache) == 2
    assert h1 != h2


def test_string_counter2():
    sc = StringCounter()
    sc['hi'] += 1
    assert isinstance(sc, StringCounter)
    sc += Counter('hi')
    assert isinstance(sc, StringCounter)


def test_unicode():
    fc = FeatureCollection({'NAME': {'foo': 1}})
    fc = FeatureCollection.loads(fc.dumps())
    assert type(fc.keys()[0]) is unicode
    assert type(fc['NAME'].keys()[0]) is unicode



'''dossier.fc Feature Collections

.. This software is released under an MIT/X11 open source license.
   Copyright 2012-2014 Diffeo, Inc.

'''
from operator import itemgetter
import sys

from dossier.fc import StringCounter, FeatureCollectionChunk


def key_sorted_dict_str(mset):
    return ', '.join(['%s: %d' % (k, mset[k])
                      for k in sorted(mset.iterkeys())])


def repr_feature(feature, max_keys=100, indent=8, lexigraphic=False):
    '''
    generate a pretty-printed string for a feature

    Currently implemented:
      * StringCounter

    @max_keys: truncate long counters

    @indent: indent multi-line displays by this many spaces

    @lexigraphic: instead of sorting counters by count (default), sort
    keys lexigraphically
    '''
    if isinstance(feature, StringCounter):
        if lexigraphic:
            recs = list(feature.iteritems())
            recs.sort(key=itemgetter(0))
            recs = recs[:max_keys]
        else:
            recs = feature.most_common(max_keys)
    elif isinstance(feature, unicode):
        return feature
    else:
        raise Exception('repr_feature(%r) not yet implemented' % type(feature))

    joiner = ', '
    if isinstance(feature, StringCounter):
        ## might have very long keys, e.g. whole sentences or URLs
        longest_key = max(map(len, map(itemgetter(0), recs)))
        if longest_key > 20:
            recs = map(lambda x: '%d: %r' % (x[1], x[0]), recs)
            joiner = '\n' + ' ' * indent
        else:
            recs = map(lambda x: '%r: %d' % (x[0], x[1]), recs)
            joiner = ', '

    len_feature = len(feature)
    if len_feature > max_keys:
        recs.append(' truncated at %d' % max_keys)
    recs.insert(0, ' (%d keys)' % len_feature)
    return joiner.join(recs)


def detailed_view(fc, features_to_show=None, max_keys=100, lexigraphic=False):
    '''
    returns a pretty-printed string describing a
    :class:`dossier.fc.FeatureCollection`.
    
    @features_to_show: list of features to include, defaults to None
    meaning include all.

    @lexigraphic: passed to repr_feature
    '''
    if 'NAME' in fc:
        top_lines = ['NAME: %r' % key_sorted_dict_str( fc.get('NAME') )]
    elif 'NOM' in fc:
        top_lines = ['NOM: %r' % key_sorted_dict_str( fc.get('NOM') )]
    else:
        top_lines = ['no-NAME, no-NOM']
    bottom_lines = []
    feature_names = fc.keys()
    feature_names.sort()
    for key in feature_names:
        if key in fc and fc[key]:
            if features_to_show and key not in features_to_show:
                continue
            repr = repr_feature(fc[key], max_keys=max_keys, lexigraphic=lexigraphic)
            if '\n' in repr:
                ## save these for the end
                # use four spaces, smaller than indent=8 above
                bottom_lines.append( '    %s: %s' % (key, repr) )
            else:
                top_lines.append( '    %s: %s' % (key, repr))
    return '\n'.join(top_lines + bottom_lines)


def only_specific_multisets(ent, multisets_to_show):
    '''
    returns a pretty-printed string for specific features in a FeatureCollection
    '''
    out_str = []
    for mset_name in multisets_to_show:
        for key, count in ent[mset_name].items():
            out_str.append( '%s - %d: %s' % (mset_name, count, key) )
    return '\n'.join(out_str)


def main():
    import argparse
    parser = argparse.ArgumentParser(
        description='Shows the features for each feature FeatureCollection in a Chunk',
        usage='dossier_feature_dump file.fc')
    parser.add_argument(
        'input', metavar='file.fb', nargs='?', default='-',
        help='defaults to stdin')
    parser.add_argument(
        '-o', '--out', default=None,
        dest='output', help='file to write output to, default stdout')
    parser.add_argument(
        '-f', '--feature', metavar='FEATURE_NAME', default=[],
        dest='features_to_show', action='append', help='only print these features')
    parser.add_argument(
        '--column-view', default=False, action='store_true', 
        help='include this multiset in print out')
    parser.add_argument(
        '--max-keys', default=100, type=int, 
        help='number of keys to show in counters')
    parser.add_argument(
        '--lexigraphic', default=False, action='store_true', 
        help='sort counter keys lexigraphically instead of by count')
    parser.add_argument(
        '--limit', default=None, type=int, metavar='LIMIT',
        help='stop after reading LIMIT records')
    args = parser.parse_args()

    if args.output:
        out = open(args.output, 'w')
    else:
        out = sys.stdout

    if args.input == '-':
        i_chunk = FeatureCollectionChunk(file_obj=sys.stdin)
    else:
        ## this allows Chunk to detect .xz and .gpg
        i_chunk = FeatureCollectionChunk(args.input)

    if  args.features_to_show:
        args.features_to_show = set(args.features_to_show)

    count = 0
    for fc in i_chunk:
        if args.column_view:
            out_str = only_specific_multisets(fc, args.features_to_show)
        else:
            out_str = detailed_view(
                fc, features_to_show=args.features_to_show, 
                max_keys=args.max_keys, lexigraphic=args.lexigraphic)
        if out_str:
            out.write(out_str.encode('utf-8'))
            out.write('\n')
        count += 1
        if (args.limit is not None) and (count >= args.limit):
            break

if __name__ == '__main__':
    main()



'''Collections of named features.

This module provides :class:`dossier.fc.FeatureCollection` and a
number of supporting classes.  A feature collection is a dictionary
mapping a feature name to a feature representation.  The
representation is typically a :class:`dossier.fc.StringCounter`, an
implementation of :class:`collections.Counter`, which is fundamentally
a mapping from a string value to an integer.

This representation allows multiple values to be stored for multiple
bits of information about some entity of interest.  The weights in the
underlying counters can be used to indicate how common a particular
value is, or how many times it appears in the source documents.

    president = FeatureCollection()
    president['NAME']['Barack Obama'] += 1
    president['entity_type']['PER'] += 1
    president['PER_ADDRESS']['White House'] += 1
    president['PER_ADDRESS']['1600 Pennsylvania Ave.'] += 1

Feature collections can have representations other than the basic
string-counter representation; these representations may not preserve
the source strings, but will still be suitable for machine-learning
applications.

Feature collections can be serialized to :rfc:`7049` CBOR format,
similar to a binary JSON representation.  They can also be stored
sequentially in flat files using
:class:`dossier.fc.FeatureCollectionChunk` as an accessor.

.. autoclass:: FeatureCollection
   :show-inheritance:

.. autoclass:: StringCounter
   :show-inheritance:

.. autoclass:: FeatureCollectionChunk
   :show-inheritance:

.. autoclass:: ReadOnlyException
   :show-inheritance:

.. autoclass:: SerializationError
   :show-inheritance:

.. Your use of this software is governed by your license agreement.
.. This software is released under an MIT/X11 open source license.
   Copyright 2012-2014 Diffeo, Inc.

'''
from dossier.fc.feature_collection import \
    FeatureCollection, FeatureCollectionChunk
from dossier.fc.string_counter import StringCounter
from dossier.fc.exceptions import ReadOnlyException, SerializationError

__all__ = [
    'FeatureCollection', 'FeatureCollectionChunk',
    'StringCounter',
    'ReadOnlyException', 'SerializationError',
]



'''dossier.fc Feature Collections

.. This software is released under an MIT/X11 open source license.
   Copyright 2012-2014 Diffeo, Inc.

'''

from __future__ import absolute_import, division, print_function

import pytest

from dossier.fc.feature_collection import registry, FeatureTypeRegistry


def is_testable_counter(k):
    return 'Counter' in k or k == 'SparseVector'


@pytest.yield_fixture(params=filter(is_testable_counter, registry.types()))
def counter_type(request):
    ct = registry.get_constructor(request.param)
    old_default = FeatureTypeRegistry.DEFAULT_FEATURE_TYPE_NAME
    FeatureTypeRegistry.DEFAULT_FEATURE_TYPE_NAME = request.param
    yield ct
    FeatureTypeRegistry.DEFAULT_FEATURE_TYPE_NAME = old_default



''':mod:`dossier.fc` exceptions.

.. This software is released under an MIT/X11 open source license.
   Copyright 2012-2014 Diffeo, Inc.

.. autoclass:: ReadOnlyException
.. autoclass:: SerializationError

'''

class BaseException(Exception):
    '''Base exception for all things in dossier.fc
    '''
    pass


class ReadOnlyException(BaseException):
    '''Code attempted to modify a read-only feature collection.

    This occurs when adding, deleting, or making other in-place
    modifications to a :class:`~dossier.fc.FeatureCollection` that has
    its :attr:`~dossier.fc.FeatureCollection.read_only` flag set.  It
    also occurs when attempting to make changes to a
    :class:`~dossier.fc.StringCounter` contained in such a collection.

    '''
    pass


class SerializationError(BaseException):
    '''A problem occurred serializing or deserializing.

    This can occur if a :class:`~dossier.fc.FeatureCollection` has an
    unrecognized feature type, or if a CBOR input does not have the
    correct format.

    '''
    pass
